@@ -1,3 +1,6 @@
+// SPDX-FileCopyrightText: 2023 The Pion community <https://pion.ly>
+// SPDX-License-Identifier: MIT
+
 package rtp
 
 import (
@@ -6,13 +9,13 @@ import (
 	"io"
 )
 
-// Extension RTP Header extension
+// Extension RTP Header extension.
 type Extension struct {
 	id      uint8
 	payload []byte
 }
 
-// Header represents an RTP packet header
+// Header represents an RTP packet header.
 type Header struct {
 	Version          uint8
 	Padding          bool
@@ -25,41 +28,66 @@ type Header struct {
 	CSRC             []uint32
 	ExtensionProfile uint16
 	Extensions       []Extension
+
+	// PaddingLength is the length of the padding in bytes. It is not part of the RTP header
+	// (it is sent in the last byte of RTP packet padding), but logically it belongs here.
+	PaddingSize byte
+
+	// Deprecated: will be removed in a future version.
+	PayloadOffset int
 }
 
-// Packet represents an RTP Packet
+// Packet represents an RTP Packet.
 type Packet struct {
 	Header
-	Payload     []byte
-	PaddingSize byte
+	Payload []byte
+
+	PaddingSize byte // Deprecated: will be removed in a future version. Use Header.PaddingSize instead.
+
+	// Deprecated: will be removed in a future version.
+	Raw []byte
+
+	// Please do not add any new field directly to Packet struct unless you know that it is safe.
+	// pion internally passes Header and Payload separately, what causes bugs like
+	// https://github.com/pion/webrtc/issues/2403 .
 }
 
 const (
-	headerLength            = 4
-	versionShift            = 6
-	versionMask             = 0x3
-	paddingShift            = 5
-	paddingMask             = 0x1
-	extensionShift          = 4
-	extensionMask           = 0x1
-	extensionProfileOneByte = 0xBEDE
-	extensionProfileTwoByte = 0x1000
-	extensionIDReserved     = 0xF
-	ccMask                  = 0xF
-	markerShift             = 7
-	markerMask              = 0x1
-	ptMask                  = 0x7F
-	seqNumOffset            = 2
-	seqNumLength            = 2
-	timestampOffset         = 4
-	timestampLength         = 4
-	ssrcOffset              = 8
-	ssrcLength              = 4
-	csrcOffset              = 12
-	csrcLength              = 4
+	// ExtensionProfileOneByte is the RTP One Byte Header Extension Profile, defined in RFC 8285.
+	ExtensionProfileOneByte = 0xBEDE
+	// ExtensionProfileTwoByte is the RTP Two Byte Header Extension Profile, defined in RFC 8285.
+	ExtensionProfileTwoByte = 0x1000
+	// CryptexProfileOneByte is the Cryptex One Byte Header Extension Profile, defined in RFC 9335.
+	CryptexProfileOneByte = 0xC0DE
+	// CryptexProfileTwoByte is the Cryptex Two Byte Header Extension Profile, defined in RFC 9335.
+	CryptexProfileTwoByte = 0xC2DE
 )
 
-// String helps with debugging by printing packet information in a readable way
+const (
+	headerLength        = 4
+	versionShift        = 6
+	versionMask         = 0x3
+	paddingShift        = 5
+	paddingMask         = 0x1
+	extensionShift      = 4
+	extensionMask       = 0x1
+	extensionIDReserved = 0xF
+	extensionIDPadding  = 0x0
+	ccMask              = 0xF
+	markerShift         = 7
+	markerMask          = 0x1
+	ptMask              = 0x7F
+	seqNumOffset        = 2
+	seqNumLength        = 2
+	timestampOffset     = 4
+	timestampLength     = 4
+	ssrcOffset          = 8
+	ssrcLength          = 4
+	csrcOffset          = 12
+	csrcLength          = 4
+)
+
+// String helps with debugging by printing packet information in a readable way.
 func (p Packet) String() string {
 	out := "RTP PACKET:\n"
 
@@ -76,7 +104,7 @@ func (p Packet) String() string {
 
 // Unmarshal parses the passed byte slice and stores the result in the Header.
 // It returns the number of bytes read n and any error.
-func (h *Header) Unmarshal(buf []byte) (n int, err error) { //nolint:gocognit
+func (h *Header) Unmarshal(buf []byte) (n int, err error) { //nolint:gocognit,cyclop
 	if len(buf) < headerLength {
 		return 0, fmt.Errorf("%w: %d < %d", errHeaderSizeInsufficient, len(buf), headerLength)
 	}
@@ -100,7 +128,7 @@ func (h *Header) Unmarshal(buf []byte) (n int, err error) { //nolint:gocognit
 	h.Padding = (buf[0] >> paddingShift & paddingMask) > 0
 	h.Extension = (buf[0] >> extensionShift & extensionMask) > 0
 	nCSRC := int(buf[0] & ccMask)
-	if cap(h.CSRC) < nCSRC || h.CSRC == nil {
+	if cap(h.CSRC) < nCSRC {
 		h.CSRC = make([]uint32, nCSRC)
 	} else {
 		h.CSRC = h.CSRC[:nCSRC]
@@ -111,6 +139,7 @@ func (h *Header) Unmarshal(buf []byte) (n int, err error) { //nolint:gocognit
 		return n, fmt.Errorf("size %d < %d: %w", len(buf), n,
 			errHeaderSizeInsufficient)
 	}
+	headerLength := n
 
 	h.Marker = (buf[1] >> markerShift & markerMask) > 0
 	h.PayloadType = buf[1] & ptMask
@@ -124,11 +153,9 @@ func (h *Header) Unmarshal(buf []byte) (n int, err error) { //nolint:gocognit
 		h.CSRC[i] = binary.BigEndian.Uint32(buf[offset:])
 	}
 
-	if h.Extensions != nil {
-		h.Extensions = h.Extensions[:0]
-	}
+	h.Extensions = h.Extensions[:0]
 
-	if h.Extension {
+	if h.Extension { // nolint: nestif
 		if expected := n + 4; len(buf) < expected {
 			return n, fmt.Errorf("size %d < %d: %w",
 				len(buf), expected,
@@ -140,69 +167,63 @@ func (h *Header) Unmarshal(buf []byte) (n int, err error) { //nolint:gocognit
 		n += 2
 		extensionLength := int(binary.BigEndian.Uint16(buf[n:])) * 4
 		n += 2
+		extensionEnd := n + extensionLength
+		headerLength = extensionEnd
 
-		if expected := n + extensionLength; len(buf) < expected {
-			return n, fmt.Errorf("size %d < %d: %w",
-				len(buf), expected,
-				errHeaderSizeInsufficientForExtension,
-			)
+		if len(buf) < extensionEnd {
+			return n, fmt.Errorf("size %d < %d: %w", len(buf), extensionEnd, errHeaderSizeInsufficientForExtension)
 		}
 
-		switch h.ExtensionProfile {
-		// RFC 8285 RTP One Byte Header Extension
-		case extensionProfileOneByte:
-			end := n + extensionLength
-			for n < end {
-				if buf[n] == 0x00 { // padding
+		if h.ExtensionProfile == ExtensionProfileOneByte || h.ExtensionProfile == ExtensionProfileTwoByte {
+			var (
+				extid      uint8
+				payloadLen int
+			)
+
+			for n < extensionEnd {
+				if buf[n] == extensionIDPadding { // padding
 					n++
+
 					continue
 				}
 
-				extid := buf[n] >> 4
-				len := int(buf[n]&^0xF0 + 1)
-				n++
+				if h.ExtensionProfile == ExtensionProfileOneByte {
+					extid = buf[n] >> 4
+					payloadLen = int(buf[n]&^0xF0 + 1)
+					n++
 
-				if extid == extensionIDReserved {
-					break
-				}
+					// Stop parsing extensions if we reach the reserved ID or padding with non-zero length
+					if extid == extensionIDReserved || extid == extensionIDPadding {
+						break
+					}
+				} else {
+					extid = buf[n]
+					n++
 
-				extension := Extension{id: extid, payload: buf[n : n+len]}
-				h.Extensions = append(h.Extensions, extension)
-				n += len
-			}
+					if extensionEnd <= n {
+						return n, fmt.Errorf("size %d < %d: %w", extensionEnd, n, errHeaderSizeInsufficientForExtension)
+					}
 
-		// RFC 8285 RTP Two Byte Header Extension
-		case extensionProfileTwoByte:
-			end := n + extensionLength
-			for n < end {
-				if buf[n] == 0x00 { // padding
+					payloadLen = int(buf[n])
 					n++
-					continue
 				}
 
-				extid := buf[n]
-				n++
+				if extensionPayloadEnd := n + payloadLen; extensionEnd < extensionPayloadEnd {
+					return n, fmt.Errorf("size %d < %d: %w", extensionEnd, extensionPayloadEnd, errHeaderSizeInsufficientForExtension)
+				}
 
-				len := int(buf[n])
-				n++
-
-				extension := Extension{id: extid, payload: buf[n : n+len]}
+				extension := Extension{id: extid, payload: buf[n : n+payloadLen]}
 				h.Extensions = append(h.Extensions, extension)
-				n += len
+				n += payloadLen
 			}
-
-		default: // RFC3550 Extension
-			if len(buf) < n+extensionLength {
-				return n, fmt.Errorf("%w: %d < %d",
-					errHeaderSizeInsufficientForExtension, len(buf), n+extensionLength)
-			}
-
-			extension := Extension{id: 0, payload: buf[n : n+extensionLength]}
+		} else {
+			// RFC3550 Extension
+			extension := Extension{id: 0, payload: buf[n:extensionEnd]}
 			h.Extensions = append(h.Extensions, extension)
-			n += len(h.Extensions[0].payload)
 		}
 	}
-	return n, nil
+
+	return headerLength, nil
 }
 
 // Unmarshal parses the passed byte slice and stores the result in the Packet.
@@ -211,15 +232,24 @@ func (p *Packet) Unmarshal(buf []byte) error {
 	if err != nil {
 		return err
 	}
+
 	end := len(buf)
 	if p.Header.Padding {
-		p.PaddingSize = buf[end-1]
-		end -= int(p.PaddingSize)
+		if end <= n {
+			return errTooSmall
+		}
+		p.Header.PaddingSize = buf[end-1]
+		end -= int(p.Header.PaddingSize)
+	} else {
+		p.Header.PaddingSize = 0
 	}
+	p.PaddingSize = p.Header.PaddingSize
 	if end < n {
 		return errTooSmall
 	}
+
 	p.Payload = buf[n:end]
+
 	return nil
 }
 
@@ -231,11 +261,12 @@ func (h Header) Marshal() (buf []byte, err error) {
 	if err != nil {
 		return nil, err
 	}
+
 	return buf[:n], nil
 }
 
 // MarshalTo serializes the header and writes to the buffer.
-func (h Header) MarshalTo(buf []byte) (n int, err error) {
+func (h Header) MarshalTo(buf []byte) (n int, err error) { //nolint:cyclop
 	/*
 	 *  0                   1                   2                   3
 	 *  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
@@ -258,7 +289,7 @@ func (h Header) MarshalTo(buf []byte) (n int, err error) {
 
 	// The first byte contains the version, padding bit, extension bit,
 	// and csrc size.
-	buf[0] = (h.Version << versionShift) | uint8(len(h.CSRC))
+	buf[0] = (h.Version << versionShift) | uint8(len(h.CSRC)) // nolint: gosec // G115
 	if h.Padding {
 		buf[0] |= 1 << paddingShift
 	}
@@ -291,34 +322,39 @@ func (h Header) MarshalTo(buf []byte) (n int, err error) {
 
 		switch h.ExtensionProfile {
 		// RFC 8285 RTP One Byte Header Extension
-		case extensionProfileOneByte:
+		case ExtensionProfileOneByte:
 			for _, extension := range h.Extensions {
-				buf[n] = extension.id<<4 | (uint8(len(extension.payload)) - 1)
+				buf[n] = extension.id<<4 | (uint8(len(extension.payload)) - 1) // nolint: gosec // G115
 				n++
 				n += copy(buf[n:], extension.payload)
 			}
 		// RFC 8285 RTP Two Byte Header Extension
-		case extensionProfileTwoByte:
+		case ExtensionProfileTwoByte:
 			for _, extension := range h.Extensions {
 				buf[n] = extension.id
 				n++
-				buf[n] = uint8(len(extension.payload))
+				buf[n] = uint8(len(extension.payload)) // nolint: gosec // G115
 				n++
 				n += copy(buf[n:], extension.payload)
 			}
 		default: // RFC3550 Extension
-			extlen := len(h.Extensions[0].payload)
-			if extlen%4 != 0 {
-				// the payload must be in 32-bit words.
-				return 0, io.ErrShortBuffer
+			// Zero length extension is valid per the RFC3550 spec
+			// https://www.rfc-editor.org/rfc/rfc3550#section-5.3.1
+			if len(h.Extensions) > 0 {
+				extlen := len(h.Extensions[0].payload)
+				if extlen%4 != 0 {
+					// the payload must be in 32-bit words.
+					return 0, io.ErrShortBuffer
+				}
+				n += copy(buf[n:], h.Extensions[0].payload)
 			}
-			n += copy(buf[n:], h.Extensions[0].payload)
 		}
 
 		// calculate extensions size and round to 4 bytes boundaries
 		extSize := n - startExtensionsPos
 		roundedExtSize := ((extSize + 3) / 4) * 4
 
+		// nolint: gosec // G115 false positive
 		binary.BigEndian.PutUint16(buf[extHeaderPos+2:extHeaderPos+4], uint16(roundedExtSize/4))
 
 		// add padding to reach 4 bytes boundaries
@@ -341,17 +377,19 @@ func (h Header) MarshalSize() int {
 
 		switch h.ExtensionProfile {
 		// RFC 8285 RTP One Byte Header Extension
-		case extensionProfileOneByte:
+		case ExtensionProfileOneByte:
 			for _, extension := range h.Extensions {
 				extSize += 1 + len(extension.payload)
 			}
 		// RFC 8285 RTP Two Byte Header Extension
-		case extensionProfileTwoByte:
+		case ExtensionProfileTwoByte:
 			for _, extension := range h.Extensions {
 				extSize += 2 + len(extension.payload)
 			}
 		default:
-			extSize += len(h.Extensions[0].payload)
+			if len(h.Extensions) > 0 {
+				extSize += len(h.Extensions[0].payload)
+			}
 		}
 
 		// extensions size must have 4 bytes boundaries
@@ -361,58 +399,68 @@ func (h Header) MarshalSize() int {
 	return size
 }
 
-// SetExtension sets an RTP header extension
-func (h *Header) SetExtension(id uint8, payload []byte) error { //nolint:gocognit
-	if h.Extension {
-		switch h.ExtensionProfile {
-		// RFC 8285 RTP One Byte Header Extension
-		case extensionProfileOneByte:
-			if id < 1 || id > 14 {
-				return fmt.Errorf("%w actual(%d)", errRFC8285OneByteHeaderIDRange, id)
-			}
-			if len(payload) > 16 {
-				return fmt.Errorf("%w actual(%d)", errRFC8285OneByteHeaderSize, len(payload))
-			}
-		// RFC 8285 RTP Two Byte Header Extension
-		case extensionProfileTwoByte:
-			if id < 1 || id > 255 {
-				return fmt.Errorf("%w actual(%d)", errRFC8285TwoByteHeaderIDRange, id)
-			}
-			if len(payload) > 255 {
-				return fmt.Errorf("%w actual(%d)", errRFC8285TwoByteHeaderSize, len(payload))
-			}
-		default: // RFC3550 Extension
-			if id != 0 {
-				return fmt.Errorf("%w actual(%d)", errRFC3550HeaderIDRange, id)
-			}
+// SetExtension sets an RTP header extension.
+func (h *Header) SetExtension(id uint8, payload []byte) error { //nolint:gocognit, cyclop
+	if h.Extension { // nolint: nestif
+		if err := headerExtensionCheck(h.ExtensionProfile, id, payload); err != nil {
+			return err
 		}
 
 		// Update existing if it exists else add new extension
 		for i, extension := range h.Extensions {
 			if extension.id == id {
 				h.Extensions[i].payload = payload
+
 				return nil
 			}
 		}
+
 		h.Extensions = append(h.Extensions, Extension{id: id, payload: payload})
+
 		return nil
 	}
 
 	// No existing header extensions
 	h.Extension = true
 
-	switch len := len(payload); {
-	case len <= 16:
-		h.ExtensionProfile = extensionProfileOneByte
-	case len > 16 && len < 256:
-		h.ExtensionProfile = extensionProfileTwoByte
+	switch payloadLen := len(payload); {
+	case payloadLen <= 16:
+		h.ExtensionProfile = ExtensionProfileOneByte
+	case payloadLen > 16 && payloadLen < 256:
+		h.ExtensionProfile = ExtensionProfileTwoByte
 	}
 
 	h.Extensions = append(h.Extensions, Extension{id: id, payload: payload})
+
 	return nil
 }
 
-// GetExtensionIDs returns an extension id array
+// SetExtensionWithProfile sets an RTP header extension and converts Header Extension Profile if needed.
+func (h *Header) SetExtensionWithProfile(id uint8, payload []byte, intendedProfile uint16) error {
+	if !h.Extension || h.ExtensionProfile == intendedProfile {
+		return h.SetExtension(id, payload)
+	}
+
+	// Don't mutate the packet if Set is going to fail anyway
+	if err := headerExtensionCheck(intendedProfile, id, payload); err != nil {
+		return err
+	}
+
+	// If downgrading assert that existing Extensions will work
+	if intendedProfile == ExtensionProfileOneByte {
+		for i := range h.Extensions {
+			if err := headerExtensionCheck(intendedProfile, h.Extensions[i].id, h.Extensions[i].payload); err != nil {
+				return err
+			}
+		}
+	}
+
+	h.ExtensionProfile = intendedProfile
+
+	return h.SetExtension(id, payload)
+}
+
+// GetExtensionIDs returns an extension id array.
 func (h *Header) GetExtensionIDs() []uint8 {
 	if !h.Extension {
 		return nil
@@ -426,10 +474,11 @@ func (h *Header) GetExtensionIDs() []uint8 {
 	for _, extension := range h.Extensions {
 		ids = append(ids, extension.id)
 	}
+
 	return ids
 }
 
-// GetExtension returns an RTP header extension
+// GetExtension returns an RTP header extension.
 func (h *Header) GetExtension(id uint8) []byte {
 	if !h.Extension {
 		return nil
@@ -439,10 +488,11 @@ func (h *Header) GetExtension(id uint8) []byte {
 			return extension.payload
 		}
 	}
+
 	return nil
 }
 
-// DelExtension Removes an RTP Header extension
+// DelExtension Removes an RTP Header extension.
 func (h *Header) DelExtension(id uint8) error {
 	if !h.Extension {
 		return errHeaderExtensionsNotEnabled
@@ -450,9 +500,11 @@ func (h *Header) DelExtension(id uint8) error {
 	for i, extension := range h.Extensions {
 		if extension.id == id {
 			h.Extensions = append(h.Extensions[:i], h.Extensions[i+1:]...)
+
 			return nil
 		}
 	}
+
 	return errHeaderExtensionNotFound
 }
 
@@ -469,29 +521,38 @@ func (p Packet) Marshal() (buf []byte, err error) {
 }
 
 // MarshalTo serializes the packet and writes to the buffer.
-func (p Packet) MarshalTo(buf []byte) (n int, err error) {
-	p.Header.Padding = p.PaddingSize != 0
+func (p *Packet) MarshalTo(buf []byte) (n int, err error) {
+	if p.Header.Padding && p.paddingSize() == 0 {
+		return 0, errInvalidRTPPadding
+	}
+
 	n, err = p.Header.MarshalTo(buf)
 	if err != nil {
 		return 0, err
 	}
 
+	return marshalPayloadAndPaddingTo(buf, n, &p.Header, p.Payload, p.paddingSize())
+}
+
+func marshalPayloadAndPaddingTo(buf []byte, offset int, header *Header, payload []byte, paddingSize byte,
+) (n int, err error) {
 	// Make sure the buffer is large enough to hold the packet.
-	if n+len(p.Payload)+int(p.PaddingSize) > len(buf) {
+	if offset+len(payload)+int(paddingSize) > len(buf) {
 		return 0, io.ErrShortBuffer
 	}
 
-	m := copy(buf[n:], p.Payload)
-	if p.Header.Padding {
-		buf[n+m+int(p.PaddingSize-1)] = p.PaddingSize
+	m := copy(buf[offset:], payload)
+
+	if header.Padding {
+		buf[offset+m+int(paddingSize-1)] = paddingSize
 	}
 
-	return n + m + int(p.PaddingSize), nil
+	return offset + m + int(paddingSize), nil
 }
 
 // MarshalSize returns the size of the packet once marshaled.
 func (p Packet) MarshalSize() int {
-	return p.Header.MarshalSize() + len(p.Payload) + int(p.PaddingSize)
+	return p.Header.MarshalSize() + len(p.Payload) + int(p.paddingSize())
 }
 
 // Clone returns a deep copy of p.
@@ -503,6 +564,7 @@ func (p Packet) Clone() *Packet {
 		copy(clone.Payload, p.Payload)
 	}
 	clone.PaddingSize = p.PaddingSize
+
 	return clone
 }
 
@@ -524,5 +586,48 @@ func (h Header) Clone() Header {
 		}
 		clone.Extensions = ext
 	}
+
 	return clone
 }
+
+func (p *Packet) paddingSize() byte {
+	if p.Header.PaddingSize > 0 {
+		return p.Header.PaddingSize
+	}
+
+	return p.PaddingSize
+}
+
+// MarshalPacketTo serializes the header and payload into bytes.
+// Parts of pion code passes RTP header and payload separately, so this function
+// is provided to help with that.
+//
+// Deprecated: this function is a temporary workaround and will be removed in pion/webrtc v5.
+func MarshalPacketTo(buf []byte, header *Header, payload []byte) (int, error) {
+	n, err := header.MarshalTo(buf)
+	if err != nil {
+		return 0, err
+	}
+
+	return marshalPayloadAndPaddingTo(buf, n, header, payload, header.PaddingSize)
+}
+
+// PacketMarshalSize returns the size of the header and payload once marshaled.
+// Parts of pion code passes RTP header and payload separately, so this function
+// is provided to help with that.
+//
+// Deprecated: this function is a temporary workaround and will be removed in pion/webrtc v5.
+func PacketMarshalSize(header *Header, payload []byte) int {
+	return header.MarshalSize() + len(payload) + int(header.PaddingSize)
+}
+
+// HeaderAndPacketMarshalSize returns the size of the header and full packet once marshaled.
+// Parts of pion code passes RTP header and payload separately, so this function
+// is provided to help with that.
+//
+// Deprecated: this function is a temporary workaround and will be removed in pion/webrtc v5.
+func HeaderAndPacketMarshalSize(header *Header, payload []byte) (headerSize int, packetSize int) {
+	headerSize = header.MarshalSize()
+
+	return headerSize, headerSize + len(payload) + int(header.PaddingSize)
+}
